#ifndef THREADS_HPP_BhjgkfG8
#define THREADS_HPP_BhjgkfG8

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(_WIN64)


    #include <windows.h>
    #include <process.h>

    #define THREAD_HANDLE HANDLE
    #define THREAD_RETURN_TYPE void
    #define THREAD_RETURN
    #define MUTEX HANDLE


// It's very simplified
inline int pthread_create(THREAD_HANDLE *thread, const void *ignored,
    THREAD_RETURN_TYPE (*start_routine)(void*), void *arg)
{
    uintptr_t handle = _beginthread(start_routine, 0, arg);
    if(handle == 0 || handle == (uintptr_t)-1L)
    {
        return errno;
    }
    *thread = (HANDLE)handle;
    return 0;
}
// It's very simplified
inline int pthread_join(THREAD_HANDLE thread, void **ignored)
{
    return WaitForSingleObject(thread, INFINITE) != WAIT_OBJECT_0;
}
inline MUTEX _NEW_MUTEX()
{
    return CreateMutex(0, 0, 0);

}
    #define NEW_MUTEX (_NEW_MUTEX())
inline void destroy_mutex(MUTEX *mutex)
{
    CloseHandle(*mutex);
}
inline void lock_mutex(MUTEX *mutex)
{
    WaitForSingleObject(*mutex, INFINITE);
}
inline void unlock_mutex(MUTEX *mutex)
{
    ReleaseMutex(*mutex);
}
#else

    #include <pthread.h>
    #define THREAD_HANDLE pthread_t
    #define THREAD_RETURN_TYPE void*
    #define THREAD_RETURN NULL
    #define MUTEX pthread_mutex_t
    #define NEW_MUTEX PTHREAD_MUTEX_INITIALIZER
    
inline void destroy_mutex(MUTEX *mutex)
{
    pthread_mutex_destroy(mutex);
}
 
inline void lock_mutex(MUTEX *mutex)
{
    pthread_mutex_lock(mutex);
}
inline void unlock_mutex(MUTEX *mutex)
{
    pthread_mutex_unlock(mutex);
}
    
#endif

#endif // THREADS_HPP_BhjgkfG8